
#include <QtGui>
#include <QDockWidget>
#include <QMessageBox>

#include "CorteoUi.h"
#include "CalibrationForm.h"
#include "CorteoParamWidget.h"
#include "../corteo/techniques.h"
#include "constants.h"

#include "XYQt/FSDoubleValidator.h"
#include "XYQt/colorseries.h"


// Class that contains a pointer to a calibration form
// and a method to compute channel from energy
class Calibration {

    CalibrationForm * calibrationForm;
    double zero;
    double slope;
    double quad;
    double resolution;

public:
    unsigned int maxchannel;

    Calibration();

    bool setCalibrationForm(CalibrationForm * cf, double maxEnergy);
    unsigned int channel(double E, bool *ok = NULL);
    double energy(unsigned int channel);
    double getResolution() {return resolution;}
    double getQuad() {return quad;}
    double getSlope() {return slope;}
    double getZero() {return zero;}
};

Calibration::Calibration() {
    calibrationForm = NULL;
}

bool Calibration::setCalibrationForm(CalibrationForm * cf, double maxEnergy) {
        calibrationForm = cf;
        if(calibrationForm!=NULL) {
            zero = calibrationForm->lineEditZero->text().toDouble();
            slope = calibrationForm->lineEditSlope->text().toDouble();
            quad = calibrationForm->lineEditQuad->text().toDouble();
            resolution = calibrationForm->lineEditResolution->text().toDouble();
            if(slope==0 && quad==0) return false;
            bool ok;
            maxchannel = channel(maxEnergy, &ok);
            return ok;
        }
        zero = 0;
        slope = 0;
        quad = 0;
        resolution = 0;
        maxchannel = 0;

        return false;
}

double Calibration::energy(unsigned int channel) {
    return zero+slope*channel+quad*channel*channel;
}


// compute channel for energy E
unsigned int Calibration::channel(double E, bool *ok) {
    double returnval;
    if(quad==0.0) {
        // linear calibration
        if(slope == 0) {
            if(ok)
                *ok = false;
            return 0;
        } else {
            returnval = (E-zero)/slope;
            if(returnval < 0) {
                if(ok)
                    *ok = false;
                return 0;
            }
            if(ok)
                *ok = true;
            return (unsigned int) returnval;
        }
    }

    // quadratic calibration
    double quad1 = 1/quad; // (already checked that quad!=0)
    double min = - 0.5*slope*quad1;
    returnval = sqrt(E*quad1 + min*min - zero*quad1) + min;
    if(returnval < 0) {
        if(ok)
            *ok = true;
        return 0;
    }
    return (unsigned int) returnval;
}


CorteoUi::CorteoUi(QWidget *parent, int maxlayer, int maxelement, QSplashScreen * splash)
    : QMainWindow(parent)
{

    qApp->setWindowIcon(QIcon(":/images/Corteo.png") );

    // create graphs and tab containing them
    tab = new QTabWidget(this);
    setCentralWidget(tab); // tab is the central widget

    mainGraph = new FSGraph("corteo.cgr",tab);
    angleGraph = new FSGraph("angle.cgr",tab);
    angleGraph->x->title = QString("Energy [eV]");
    angleGraph->y->title = QString("Main-scattering-to-detector angle [deg.]");
    angleGraph->setToolTip("Graph of the angle between the trajectory of the\n"
                           "detected particle at the exit of the main scattering\n"
                           "event and the detector axis as a function of detected energy.");

    tab->addTab(mainGraph, "Spectra/profiles");
    tab->addTab(angleGraph, "Angle to detector");
    tab->setTabEnabled (1, false);

    // create a parameter toolbox
    paramUi = new CorteoParamWidget(this, maxlayer, maxelement, splash);

    // initialize progress dialog pointer
    progress = NULL;

    // toolbox size defines dock size, a bit larger though
    QSize s = paramUi->toolBox->size();
    int w = s.width()+5;
    int h = s.height();
    paramUi->setFixedWidth(w+5);
    paramUi->setMinimumHeight(h);

    // create an unclosable dock widget that holds the parameter toolbox
    QDockWidget *dock = new QDockWidget("Parameters", this);
    dock->setAllowedAreas(Qt::LeftDockWidgetArea | Qt::RightDockWidgetArea);
    dock->setWidget(paramUi);
    addDockWidget(Qt::RightDockWidgetArea, dock);
    dock->setFixedWidth(w);
    dock->setFeatures(QDockWidget::DockWidgetMovable | QDockWidget::DockWidgetFloatable);

    // graph will appear square
    resize(h+w,h);

    connect( paramUi->pushButtonApplyCalib,   SIGNAL(clicked()),   SLOT(applyCalib())      );
    connect( paramUi->pushButtonGo,           SIGNAL(clicked()),   SLOT(go())              );
    connect( paramUi->comboBoxTechnique,      SIGNAL(currentIndexChanged(int)),   SLOT(UpdateConnectCorteoLoadPushbutton()) );
    connect( qApp, SIGNAL(lastWindowClosed()), paramUi,  SLOT(saveParam()));

    connect( paramUi->pushButtonSaveCorteoIn,  SIGNAL(clicked()),   SLOT(writeCorteoIn())  );
    UpdateConnectCorteoLoadPushbutton();

    corteoProcess = new QProcess( this );

}


void CorteoUi::UpdateConnectCorteoLoadPushbutton() {
    paramUi->pushButtonLoadCorteoDetect->disconnect();
    int tech = paramUi->technique();


    switch(tech) {
    case TECH_IMPLANT :
    case TECH_IMPREC :
        connect( paramUi->pushButtonLoadCorteoDetect,   SIGNAL(clicked()),   SLOT(readCorteoOut())      );
        paramUi->pushButtonLoadCorteoDetect->setToolTip(
                    tr("Load implantation simulation output\nWarning: assumes current CorteoUI parameters are correct for this file."));
      break;
    case TECH_RBS_TOF :
    case TECH_RBS :
    case TECH_ERD_TOF :
    case TECH_ERD :
    case TECH_COINC :
        connect( paramUi->pushButtonLoadCorteoDetect,   SIGNAL(clicked()),   SLOT(readCorteoDetect())      );
        paramUi->pushButtonLoadCorteoDetect->setToolTip(
                    tr("Load corteo.detect for normal simulation.\nWarning: assumes current CorteoUI parameters are correct for this file."));
        break;
    case TECH_FULLRBS :
    case TECH_FULLBACK :
        connect( paramUi->pushButtonLoadCorteoDetect,   SIGNAL(clicked()),   SLOT(readCorteoDetectFullRBS())      );
        paramUi->pushButtonLoadCorteoDetect->setToolTip(
                    tr("Load corteo.detect for full simulation.\nWarning: assumes current CorteoUI parameters are correct for this file."));
        break;
    default:
        QMessageBox::warning(this, "Corteo user interface", QString("Unknown technique code %1").arg(tech));
        break;
    }

}

// few checks and progress bar setup, then start simulation
bool CorteoUi::go() {

    // count the number of simulations to carry out
    bool isUsedElement = false;
    for(int ilayer=0; ilayer<paramUi->numberLayer() && !isUsedElement; ilayer++)
        for(int ielement=0; ielement<paramUi->numberElement(ilayer) && !isUsedElement; ielement++)
            if(paramUi->isUsedElement(ilayer, ielement))
                isUsedElement = true;

    // anything to simulate?
    if(!isUsedElement ) {
        QMessageBox::warning( this, "CorteoUi", "No 'used' element.");
        return false;
    }
    if(paramUi->lineEditCharge->text().toDouble()==0) {
        QMessageBox::warning( this, "CorteoUi", "Incident charge is 0.");
        return false;
    }
    if(paramUi->checkBoxListMode->isChecked() && paramUi->listFile.isEmpty()) {
        QMessageBox::warning( this, "CorteoUi", "List mode selected but no list file specified.");
        return false;
    }

    // anything to simulate?
    int technique = paramUi->technique();
    if( ( technique>0 || technique == TECH_FULLRBS) &&
            paramUi->numberSimul()==0) {
              QMessageBox::warning( this, "CorteoUi", "No element selected for simulation.\n(Check 'Simulate this element' boxes in layers.)");
              return false;
    }

    // prepare for simulations: disable 'go' button and create a progress bar
    paramUi->pushButtonGo->setEnabled(false);
    progress = new QProgressDialog(this);
    progress->setCancelButton ( new QPushButton ("Cancel", progress) );

    if(technique==TECH_FULLBACK) {
        // run a full RBS simulation, but upon results loading, readFullRBS() will cut-off events for which
        // the largest scattering event is ABOVE cut-off, therefore retaining only background, below cut-off events
        if(!runCorteo(TECH_FULLRBS)) {
            paramUi->pushButtonGo->setEnabled(true);
            return false;
        }
    } else {
        if(!runCorteo(technique)) {
            paramUi->pushButtonGo->setEnabled(true);
            return false;
        }
    }
    return true;
}


void CorteoUi::endSimulations() {

    progress->disconnect();
    delete progress;
    progress = NULL;

    corteoProcess->disconnect();
    corteoOutput  += corteoProcess->readAllStandardOutput();
    corteoError   += corteoProcess->readAllStandardError();

    if(corteoProcess->state()==QProcess::Running)
        corteoProcess->kill();

    printLog();

    paramUi->pushButtonGo->setEnabled(true);
}


// write input parameters to corteo.in
bool CorteoUi::writeCorteoIn(int technique) {
    // get the corteo input parameters
    bool error;
    QString s = paramUi->corteo(&error, technique);

    if(error) {
        // something's wrong, show error message contained in corteo input and stop
        QMessageBox::warning(this, "CorteoUi", s);
        return false;
    }

    QFile file( "corteo.in" );
    if( !file.open( QIODevice::WriteOnly ) ) {
        QMessageBox::warning( 0, "CorteoUi", "Can't open 'corteo.in'.");
        return false;
    }
    QTextStream ts( &file );
    ts << s << endl;
    file.close();
    return true;
}

// overload of the previous where the technique is determined from
// the value of the technique combobox
bool CorteoUi::writeCorteoIn() {
    int technique = paramUi->technique();
    if(technique==TECH_FULLBACK)
        technique = TECH_FULLRBS;
    return writeCorteoIn(technique);

}


// run a Corteo simulation
bool CorteoUi::runCorteo(int technique) {

    if(corteoProcess->state()!=QProcess::NotRunning) {
        QMessageBox::warning( 0, "CorteoUi", "Corteo already running.");
        return false;
     }

    if(writeCorteoIn(technique)==false) {
        return false;
    }

    // remove old output files so we are not confused reading old results
    if(technique>0 || technique == TECH_FULLRBS)
        QFile::remove("corteo.detect");
    else
        QFile::remove("impl_beam.out");

    // create a process to run corteo    
#ifdef _WINDOWS_
    corteoProcess->start("corteo.exe");
#else
    corteoProcess->start("./corteo");
#endif

    if(corteoProcess->waitForStarted() == false) {
        QMessageBox::warning( 0, "Corteo error", "Corteo could not start.");
        corteoProcess->kill();
        return false;
    }

    // erase content of last output, if any
    corteoOutput  = "";
    corteoError   = "";
    progress->setLabelText("Simulation starting...");

    // when corteo outputs something, call progress update
    connect(corteoProcess, SIGNAL(readyReadStandardOutput()), this, SLOT(updateProgress()));

    // when corteo finishes, call the corresponing output read function
    if(technique == TECH_FULLRBS)
        connect(corteoProcess, SIGNAL(finished(int ,QProcess::ExitStatus)),
                this, SLOT(readFullRBS(int)));
    else if(technique<0)
        connect(corteoProcess, SIGNAL(finished(int ,QProcess::ExitStatus)),
                this, SLOT(readImplOutput(int)));
    else
        connect(corteoProcess, SIGNAL(finished(int ,QProcess::ExitStatus)),
                this, SLOT(readIBAOutput(int)));

    connect(progress, SIGNAL(canceled()), this, SLOT(endSimulations()));
    progress->exec();

    return true;
}


// slot called whenever some output comes out of corteo
void CorteoUi::updateProgress() {

    // append output
    corteoOutput  += corteoProcess->readAllStandardOutput();
    corteoError   += corteoProcess->readAllStandardError();

    // figure out the number of threads and simulations
    int i = corteoOutput.indexOf("thread");
    int numthreads = i>0?corteoOutput.left(i).toInt():0;

    i = corteoOutput.indexOf("simulation");
    int numSimul = i>0?corteoOutput.mid(i-4,3).toInt():0;

    // number of "_" tels us the progress of the simulation
    int count = corteoOutput.count('_');

    // at the end of the simulation, there are 10 "_" per simulation and per  threads
    progress->setRange(0, 10*numSimul*numthreads);
    progress->setValue(count);
    progress->setLabelText(QString("Simulation %1 of %2").arg(int(count/10/numthreads)+1).arg(numSimul));

}

void CorteoUi::readImplOutput(int exitStatus) {

    corteoOutput  += corteoProcess->readAllStandardOutput();
    corteoError   += corteoProcess->readAllStandardError();

    endSimulations();

    if(exitStatus>0) {
        QMessageBox::warning(this,"Corteo","There were warnings or errors, please check log.");
        return;
    }

    readCorteoOut();
}


bool CorteoUi::readCorteoOut() {

    // when also loading displacement profile, let make implantation profile thicker
    double linewidth = paramUi->technique()==TECH_IMPREC?2:1;

    // simulation successful: create a dataset, load resulting profile into it, and add it to the graph
    FSDataSet * d = new FSDataSet( paramUi->beam().symbol, QPen(colorSeries(paramUi->beam().Z), linewidth) );
    if(d->loadfile("impl_beam.out",1,2,1)==false) { // read column 1 (depth) as x, column 2 as y, skipping line 1
        QMessageBox::warning( this, "CorteoUi", "Can't open depth profile 'impl_beam.out'. Stopping here.");
        return false;
    }

    for(int i=0; i<d->getndata(); i++)
        d->yp[i]*=1e8; // at/A/ion to [at/cm3]/[at/cm2]
    mainGraph->appendDataSet(d);
    tab->setTabEnabled (1, false);

    // in case of full cascade, also load displacements profiles
    // (but Corteo generates only one file per element, event if present in many layer, so load them only once)
    if(paramUi->technique()==TECH_IMPREC) {
        QList<int> Z;
        // loop through all layers and elements
        for(int ilayer=0; ilayer<paramUi->numberLayer(); ilayer++) {
            for(int ielement=0; ielement<paramUi->numberElement(ilayer); ielement++) {
                // if element is used
                if(paramUi->isUsedElement(ilayer, ielement)) {
                    Atom a = paramUi->element(ilayer, ielement);
                    if(!Z.contains(a.Z)) {
                        FSDataSet * d = new FSDataSet( QString("%1 disp").arg(a.symbol), QPen(colorSeries(a.Z), 1) );
                        QString filename = QString("impl%1.out").arg(a.Z,3,10,QChar('0'));
                        if(d->loadfile(filename,1,2,1)==false) { // read column (energy) 1 as x, column 2 as y, skipping line 1
                            QMessageBox::warning( this, "CorteoUi", QString("Can't open file %1. Stopping here.").arg(filename));
                            return false;
                        }
                        for(int i=0; i<d->getndata(); i++)
                            d->yp[i]*=1e8; // at/A/ion to [at/cm3]/[at/cm2] * [at/cm2]
                        mainGraph->appendDataSet(d);
                        Z.append(a.Z);
                    }
                }
            }
        }
    }

    mainGraph->y->title = QString("Concentration/fluence [at/cm^{3}]/[at/cm^{2}]");
    mainGraph->x->title = QString("Depth [A]");
    return true;

}

void CorteoUi::readFullRBS(int exitStatus) {

    corteoOutput  += corteoProcess->readAllStandardOutput();
    corteoError   += corteoProcess->readAllStandardError();

    endSimulations();

    if(exitStatus>0) {
        QMessageBox::warning(this,"Corteo","Corteo simulation ended with error. Please check log.");
        return;
    }

    readCorteoDetectFullRBS();
}


bool CorteoUi::readCorteoDetectFullRBS() {

    // simulation successful: produce a histogram as a new dataset, and add it to the graph
    QFile file("corteo.detect");
    if(!file.open(QIODevice::ReadOnly)) {
        QMessageBox::warning( this, "CorteoUi", "Can't open 'corteo.detect'.");
        return false;
    }
    bool ok;
    int numproj = paramUi->lineEditNumberProjectiles->text().toInt(&ok);
    if(!ok || numproj<=0) {
        QMessageBox::warning( this, "CorteoUi", "Please set the number of projectiles properly.");
        return false;
    }

    // each detected particle in full RBS mode contributes the indicated incident charge (in atoms)
    // divided by the number of simulated particles
    double qq = paramUi->lineEditCharge->text().toDouble()/ELEMENTARY_CHARGE/numproj;

    bool isTOF = paramUi->isTOF();
    int Z1 = paramUi->beam().Z;
    int Zdet = Z1;
    if(isTOF) Zdet = 0; // if TOF, use its calibration (form index 0)
    int technique = paramUi->technique();

    double cosCutOffAngle = paramUi->lineEditCutoff->text().toDouble(&ok);
    if(paramUi->lineEditCutoff->text().isEmpty() || !ok || cosCutOffAngle<=0 || cosCutOffAngle >= 180) {
        QMessageBox::warning( this, "CorteoUi", "Problem with cut-off angle, must be set above 0 and below 180 degrees");
        return false;
    }
    cosCutOffAngle = cos(cosCutOffAngle*PI/180.);

    PeriodicTable pt;

    // get calibration for beam
    Calibration calib;
    if(calib.setCalibrationForm(
                paramUi->getCalibrationForm(Zdet),
                paramUi->lineEditEnergy->text().toDouble()) == false) {
        QMessageBox::warning(this,
                             "CorteoUi",
                             QString("Calibration of element %1 not set properly. Correct and reload data.")
                             .arg(pt.returnAtom(Z1).symbol) );
        return false;
    }

    // create and initialize a dataset in which the histogram of detected particles will be stored
    FSDataSet * d = new FSDataSet( "Full RBS", QPen(QColor("black"), 1, Qt::DashLine) );
    for(unsigned int iE=0; iE<calib.maxchannel; iE++)
        d->append( calib.energy(iE), 0);

    FSDataSet * angleData = new FSDataSet("max scat. angle (Full RBS)", QPen(), 11, 5, QPen(QColor("magenta")));
    angleData->linePen.setWidth(0); // no line, only dots

    QTextStream in(&file);
    while(!in.atEnd()) {
        QString line = in.readLine();
        if(line[0].isNumber()) {  // comment line begin with any symbol not involve in number representation

            QStringList val = line.simplified().split(" ");

            // increment histogram (yp)
            bool ok;
            unsigned int ch = calib.channel(val[0].toDouble(), &ok); // get channel for E (column 0)
            if(ok && ( technique != TECH_FULLBACK || (technique == TECH_FULLBACK && val[1].toDouble()>cosCutOffAngle)))
                // incremeted if not background simulation or if it is and angle is below cut-off angle (i.e. cos larger)
                d->yp[ch] += qq;

            // accumulate scattering angle data
            if(paramUi->checkBoxGraphAngle->isChecked()  && val.count()>1)
                angleData->append(val[0].toDouble(),acos(val[1].toDouble())*180/PI);
        }
    }


    d->gaussConvol( calib.getResolution()/calib.getSlope() );
    mainGraph->appendDataSet(d, true);

    if(paramUi->lineEditCharge->text().toDouble()==1) {
        // charge set to 1 coulomb, normalized spectrum
        mainGraph->y->title = QString("counts/coulomb");
    } else {
        // charge set to some value
        mainGraph->y->title = QString("counts");
    }
    mainGraph->x->title = QString("Energy [eV]");


    if(angleGraph->getndatasets()>0) {
        tab-> setCurrentIndex(1);
        if(QMessageBox::question(this, "CorteoUi", "This graph contains data from previous simulation. Clear them?",
                                 QMessageBox::Yes, QMessageBox::No )==QMessageBox::Yes) {
            angleGraph->clearDataSets();
        }
    }

    addAngleData(angleData);
    appendEventsToListmodeFile();
    return true;

}


void CorteoUi::readIBAOutput(int exitStatus) {

    corteoOutput  += corteoProcess->readAllStandardOutput();
    corteoError   += corteoProcess->readAllStandardError();
    // QMessageBox::warning( 0, "CorteoUi", corteoOutput);

    endSimulations();

    if(exitStatus>0) {
        QMessageBox::warning(this,"Corteo","There were warnings or errors, please check log.");
        return;
    }

    readCorteoDetect();
}


bool CorteoUi::readCorteoDetect() {

    QFile file("corteo.detect");
    if(!file.open(QIODevice::ReadOnly)) {
        QMessageBox::warning( this, "CorteoUi", "Can't open 'corteo.detect'.");
        return false;
    }

    bool doLoadEachZ = paramUi->checkBoxLoadSeparateSpectra->isChecked();
    bool isBeam = paramUi->isBeamDetectionTechnique();   // true if the spectrum is made of scattered beam atoms
    bool isTOF = paramUi->isTOF();
    double q = paramUi->lineEditCharge->text().toDouble()/ELEMENTARY_CHARGE; // number of incident atoms
    int Z1 = paramUi->beam().Z;

    // list of Z of elements already simulated and there corresponding dataset and calibration
    QMap<int, Calibration *> ZCalibration;  // map of Calibrations for each Z
    QMap<int, FSDataSet *> ZDataSet;  // map of datasets for each Z
    PeriodicTable pt;

    FSDataSet * angleData = new FSDataSet("angle to det. at coll.", QPen(), 11, 5, QPen(QColor("blue")));
    angleData->linePen.setWidth(0);

    QTextStream in(&file);
    while(!in.atEnd()) {
        QString line = in.readLine();
        if(line[0].isNumber()) {  // comment line begin with any symbol not involve in number representation

            QStringList val = line.simplified().split(" ");
            if( val.count()<6) {
                QMessageBox::warning(this,"Corteo","Too few columns in corteo output. Expecting at least 6 columns.");
                return false;
            }

            int Z2 = val[5].toInt(); // histograms generated in different datasets in ZDataSet (index Z2)
            int Zdet = isBeam ? Z1 : Z2;  // detected particle is beam or recoil
            if(isTOF) Zdet = 0; // but if TOF, use its calibration (form index 0)

            if(ZCalibration.contains(Zdet)==false) {
                ZCalibration.insert(Zdet, new Calibration());
                if(ZCalibration[Zdet]->setCalibrationForm(
                            paramUi->getCalibrationForm(Zdet),
                            paramUi->lineEditEnergy->text().toDouble()) == false) {
                    QMessageBox::warning(this,
                                         "CorteoUi",
                                         QString("Calibration of element %1 not set properly. Correct and reload data.")
                                         .arg(pt.returnAtom(Zdet).symbol) );
                    return false;
                }
            }
            if(ZDataSet.contains(Z2)==false) {
                // create dataset to store histogram and initialize it to 0
                ZDataSet.insert(Z2, new FSDataSet(pt.returnAtom(Z2).symbol, QPen(colorSeries(Z2), 1)) );
                for(unsigned int iE=0; iE<ZCalibration[Zdet]->maxchannel; iE++)
                    ZDataSet[Z2]->append( ZCalibration[Zdet]->energy(iE), 0);
            }

            // increment histogram (yp) by cross-section * Nt * Omega / Number of detectable particles
            bool ok;
            unsigned int ch = ZCalibration[Zdet]->channel(val[0].toDouble(), &ok); // get channel for E (column 0)
            if(ok) ZDataSet[Z2]->yp[ch] += q * val[1].toDouble() * val[2].toDouble();
            // in line above, column 1 = cross section, column 2 = Nt * Omega / Number of detectable particles

            // accumulate scattering angle data
            if(paramUi->checkBoxGraphAngle->isChecked())
                angleData->append(val[0].toDouble(),acos(val[4].toDouble())*180/PI);
        }
    }

    if(ZDataSet.empty()) {
        QMessageBox::warning(this,"Corteo","Corteo output looks empty.");
        return false;
    }

    // add datasets to graph
    QMapIterator<int, FSDataSet *> i(ZDataSet); // iterator for FSDataSets in ZDataSet

    // convolution of first spectrum for resolution (i.key() = Z2)
    i.next();
    int Z2 = i.key();
    Calibration * cal = ZCalibration.value(isBeam ? Z1 : Z2 );
    i.value()->gaussConvol( cal->getResolution()/cal->getSlope() );

    // in the case of RBS, produce a total spectrum
    FSDataSet * sumSpectrum;
    if(isBeam) {
        // create a new dataset to store the sum, as a copie of the first dataset in ZDataSet
        sumSpectrum = new FSDataSet(i.value());
        sumSpectrum->legend = paramUi->beam().symbol; // legend = beam name
        sumSpectrum->linePen = QPen(colorSeries(Z1), doLoadEachZ ? 2:1 ); // thicker line of sumSpectrum is not alone
    }

    if(doLoadEachZ || !isBeam)
        mainGraph->appendDataSet(i.value(), true);

    while (i.hasNext()) {
        i.next();
        int Z2 = i.key();

        // convolution of spectrum for resolution
        cal = ZCalibration.value(isBeam ? Z1 : Z2 );
        i.value()->gaussConvol( cal->getResolution()/cal->getSlope() );

        // for RBS, add data to total spectrum
        if(isBeam)
            for (int j = 0; j<sumSpectrum->getndata(); j++)
                sumSpectrum->yp[j] += i.value()->yp[j];

        // add to graph or free memory
        if(doLoadEachZ || !isBeam)
            mainGraph->appendDataSet(i.value(), true);

    }
    // for RBS, add total spectrum to graph
    if(isBeam)
        mainGraph->appendDataSet(sumSpectrum, true);

    if(q==1) {
        // charge not set, normalized spectrum
        mainGraph->y->title = QString("counts/coulomb");
    } else {
        // charge set to some value
        mainGraph->y->title = QString("counts");
    }
    mainGraph->x->title = QString("Energy [eV]");


    if(angleGraph->getndatasets()>0) {
        tab-> setCurrentIndex(1);
        if(QMessageBox::question(this, "CorteoUi", "This graph contains data from previous simulation. Clear them?",
                                 QMessageBox::Yes, QMessageBox::No )==QMessageBox::Yes) {
            angleGraph->clearDataSets();
        }
    }

    addAngleData(angleData);
    appendEventsToListmodeFile();
    return true;

}



// if requested, append simulation results to the list-mode file
void CorteoUi::appendEventsToListmodeFile() {

    // append or replace the content of corteo.detect to the list file
    if(paramUi->checkBoxListMode->isChecked()) {
        QFile fi("corteo.detect");
        QFile fo(paramUi->listFile);

        bool append, cancel = false;

        // if list file exists, ask if new data should be appened or if the file should be replaced
        if(fo.exists()) {
            QMessageBox msgBox;
            msgBox.setText(QString("'%1' already exists").arg(paramUi->listFile));

            QPushButton *appendButton = msgBox.addButton(tr("Append"), QMessageBox::ActionRole);
            QPushButton *relaceButton = msgBox.addButton(tr("Replace"), QMessageBox::ActionRole);
            QPushButton *cancelButton = msgBox.addButton(QMessageBox::Cancel);

            msgBox.exec();

            if (msgBox.clickedButton() == appendButton) {
                append = true;
            } else if (msgBox.clickedButton() == relaceButton) {
                append = false;
            } else if (msgBox.clickedButton() == cancelButton) {
                cancel = true;
            }


        }

        if(!cancel) {
            QIODevice::OpenMode mode = QIODevice::WriteOnly | QIODevice::Text;
            if(append) mode |= QIODevice::Append;

            if(fi.open(QIODevice::ReadOnly | QIODevice::Text) && fo.open(mode)) {
                QTextStream in(&fi);
                QTextStream out(&fo);
                while (!in.atEnd()) {
                    QString line = in.readLine();
                    out << line << endl;
                }
                fi.close();
                fo.close();
            } else {
                QMessageBox::warning( this, "CorteoUi", QString("Error opening '%1' or '%2'.")
                                      .arg(fi.fileName()).arg(fo.fileName()));
            }
        }

    }

}

// if user requested an angle-to-detector graph, append dataset to the graph and show it
// also show a line delimiting 99% of the largest angles
void CorteoUi::addAngleData(FSDataSet * angleData) {

    if(paramUi->checkBoxGraphAngle->isChecked()) {

        // append angle datasets
        angleGraph->appendDataSet(angleData, true);

        // generate a curve delimiting the 99th percentile of angles to detector
        // we will generate 10 such points, each for 1/10th of the data
        // as we need the 99th percentile 10 times, we need at least 1000 points
        angleData->sort(); // in depth order
        int n = angleData->getndata()/10;
        FSDataSet * percentile99 = new FSDataSet(QString("%1 99%").arg(angleData->getname()),
                                                 QPen(QColor("red"), 2));
        if(n>100) {
            FSDataSet * temp = new FSDataSet();
            temp->grow(n);
            double xmoy = 0;
            for(int j=0; j<10; j++) {
                temp->clear();
                // create a dataset containing a subset (1/10) of the points
                for(int k=j*n; k<(j+1)*n; k++) {
                    temp->append(angleData->xp[k],angleData->yp[k]);
                    xmoy += angleData->xp[k];
                }
                xmoy /= n;

                // swap x and y so we can use the sort function on y data
                temp->swap();
                temp->sort();

                // add the point that is the 99th/100 of the sorted data
                percentile99->append(xmoy,temp->xp[int(temp->getndata()*0.99+.5)]);
            }
            angleGraph->appendDataSet(percentile99, true);
        }

        tab->setTabEnabled(1, true);
    } else {
        // no angle data, disable tab
        if(angleGraph->getndatasets()==0) {
            tab-> setCurrentIndex(0);
            tab->setTabEnabled(1, false);
        }
    }

}

// Apply currently displayed calibration to currently selected dataset
// (so data loaded in channels can be converted to energy)
bool CorteoUi::applyCalib() {
    FSDataSet * s = mainGraph->selectedDataSet;
    if(s==NULL) return false; // no currently selected dataset

    // get current calibration
    CalibrationForm * currentCal = (CalibrationForm *)paramUi->tabWidgetCalibrations->widget(paramUi->tabWidgetCalibrations->currentIndex());
    double zero =  currentCal->lineEditZero->text().toDouble();
    double slope = currentCal->lineEditSlope->text().toDouble();
    double quad =  currentCal->lineEditQuad->text().toDouble();

    if(slope==0 && quad==0) return false; // applying the calibration with only zero defined would put all channels to zero

    int i, n = s->getndata();
    for(i=0;i<n; i++) s->xp[i] = zero + s->xp[i]*slope + s->xp[i]*s->xp[i]*quad;

    mainGraph->updateGraph();
    return true;
}


bool CorteoUi::printLog() {
    // open log file
    QFile logfile("CorteoUi.log");
    if (!logfile.open(QIODevice::WriteOnly | QIODevice::Text | QIODevice::Append)) {
        QMessageBox::warning( this, "CorteoUi", "Can't open 'CorteoUi.log'.");
        return false;
    }

    QTextStream logstream(&logfile);
    QDateTime dt = QDateTime::currentDateTime();
    logstream << endl << endl << "======= " << dt.toString("yyyy/MM/dd hh:mm:ss") << " =======" << endl;
    logstream << corteoOutput << endl;
    if(!corteoError.isEmpty()) {
        logstream << "--- Warning and Error(s) ---" << endl;
        logstream << corteoError << endl ;
    }
    logstream << endl;
    logfile.close();

    return true;
}



