
/* This file is part of the Allegria project
   Copyright (C) 2004 Francois Schiettekatte <francois.schiettekatte@umontreal.ca>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

/****************************************************************************
** This class contains data and the method to draw them using a QPainter
** object against x & y axes defined by FSAxis class
** It also holds the adress of the previous and next dataset to be plotted
** Namely takes care of loading, saving, interpolations and sorting
*****************************************************************************/

#include <QFile>
#include <QTextStream>
//#include <stdlib.h>
#include <QRegExp>
#include <QPolygon>
#include <QMessageBox>

#include "FSDataSet.h"
#include "FSEditDataSet.h"

//
// Construct a FSDataSet without parameters
//
FSDataSet::FSDataSet()
{
    _init();

    legend = QString("<no name>");
    linePen = QPen(Qt::black, 1);
    symbol = 0;
    symbolSize = 10.;
    symbolPen = QPen(Qt::SolidLine);
}

//
// Construct a FSDataSet with parameters
//
FSDataSet::FSDataSet(QString llegend, QPen llinePen, int ssymbol, double ssymbolSize, QPen ssymbolPen)
{
    _init();

    legend = llegend;
    linePen = llinePen;
    symbol = ssymbol;
    symbolSize = ssymbolSize;
    symbolPen = ssymbolPen;
}

//
// Construct a FSDataSet from another one (the previous and next dataset are not copied)
//
FSDataSet::FSDataSet(FSDataSet * d)
{
    _init();

    legend     = d->legend;
    linePen    = d->linePen;
    symbolPen  = d->symbolPen;
    symbol     = d->getsymbol();
    symbolSize = d->getsymbolSize();


    int n = d->getndata();
    grow(n);

    for(int i=0 ; i<n; i++)
        append(d->xp[i], d->yp[i]);

	extrema();

    xMin = getxMin();
    xMax = getxMax();
    yMin = getyMin();
    yMax = getyMax();
    xMingt0 = getxMingt0();
    yMingt0 = getyMingt0();

}

//
// class destructor
//
FSDataSet::~FSDataSet()
{
    clear();

    if(prevDataSet) prevDataSet->nextDataSet = nextDataSet;  // tell previous dataset that now its next dataset is what was our next dataset
    if(nextDataSet) nextDataSet->prevDataSet = prevDataSet;  // ... and inversely
}



void FSDataSet::_init()
{
    ndata=0;
    maxdata=0;

    xp = NULL;
    yp = NULL;

    legendRect = QRect();
    dataHaveChanged = false;

    prevDataSet = NULL;
    nextDataSet = NULL;
}

//
// append new data
//
void FSDataSet::append(double xx, double yy)
{

    // if necessary, create larger xp,yp arrays and copy the data into them
    if (ndata+1>maxdata)
		grow(FSDATAGROW);

    // append the new data
    xp[ndata]=xx;
    yp[ndata]=yy;
    ndata++;

}


//
// just grow the available space by n
//
void FSDataSet::grow(int n)
{
    // if necessary, create larger xp,yp arrays and copy the data into them
    if (ndata+n>maxdata) {
        maxdata=(ndata+n)-(ndata+n)%FSDATAGROW+FSDATAGROW;
		double * newxp = new double[maxdata];
		double * newyp = new double[maxdata];
        for (int i=0; i<ndata; i++)
        {
            newxp[i] = xp[i];
            newyp[i] = yp[i];
        }
        if(xp) delete[] xp;
        if(yp) delete[] yp;
        xp = newxp;
        yp = newyp;
    }

/*    // put the new data to 0	-- THIS IS WRONG!! we only want allocated space
    for (int i=0; i<n; i++) {
        xp[ndata]=0.;
        yp[ndata]=0.;
        ndata++;
    }
*/
}


//
// clear data
//
void FSDataSet::clear()
{
    if(xp) delete[] xp;
    if(yp) delete [] yp;
    xp = NULL;
    yp = NULL;
    maxdata = 0;
    ndata = 0;
}


//
// load data from ASCII file with columns, ignore the rest of the line
//
bool FSDataSet::loadfile(QString filename, int colx, int coly, int ignoreFirstNlines)
{
    QFile file(filename);
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) return false;

    QTextStream f(&file);
    int maxcol = coly>colx?coly:colx;
    double xx=0., yy=0.;

    // ignore first 'ignoreFirstNlines' lines
    for(int i=0; i<ignoreFirstNlines; i++) f.readLine();

    while(!f.atEnd())
    {
        for(int i=1; i<=maxcol; i++)
        {
            QString ss;
            f >> ss;
            if(i==colx) xx = ss.toDouble();
            if(i==coly) yy = ss.toDouble();
        }
        f.readLine();

        append(xx,yy);
        if(colx<=0) xx++;
        if(coly<=0) yy++;
    }
    ndata--;
    file.close();
    return true;
}


//
// read data just as they come from an ASCII file and put them in y. x[i]=i
//
bool FSDataSet::swallow(QString filename)
{
    QFile file(filename);
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) return false;

    QTextStream f(&file);
    double xx=0., yy;

    while(!f.atEnd())
    {
        f >> yy;
        append(xx,yy);
        xx++;
    }
    file.close();
    return true;
}



//
// calculate dataset extrema
// returns true if xp[i] are in increasing order, false otherwise
//
bool FSDataSet::extrema()
{
	int i;
    bool isOrdered = true;

    if(ndata==0) // if their's no data
    {
        xMin = .1; yMin = .1;   // set defaults
        xMingt0 = .1; yMingt0 = .1;   // set defaults
	xMax = 10.; yMax = 10.;
	return isOrdered;
    }

    // initialize using first value
    xMin=xp[0]; xMax=xp[0];
    yMin=yp[0]; yMax=yp[0];
    for(i=1; i<ndata; i++) // all other data
    {
		if(xp[i] > xMax) xMax=xp[i];
        if(xp[i] < xMin) xMin=xp[i];
		if(yp[i] > yMax) yMax=yp[i];
		if(yp[i] < yMin) yMin=yp[i];

        if(xp[i] < xp[i-1]) isOrdered = false;
    }

    // now we need the minimum above 0 just in case we are in log axis mode
    xMingt0=xMax>0.?xMax:.1; 	// dummy value .1 in case all data <= 0
    yMingt0=yMax>0.?yMax:.1;
    for(i=0; i<ndata; i++) // all other data
    {
        if(xp[i] < xMingt0 && xp[i] > 0.) xMingt0=xp[i];
        if(yp[i] < yMingt0 && yp[i] > 0.) yMingt0=yp[i];
    }

    return isOrdered;
}

//
// draw the data on dc using xax and yax world transform for x and y real-to-logic coordinates transform
//
void FSDataSet::draw(QPainter *dc, FSAxis *xax, FSAxis *yax, double fSize, QRect r )
{
    QPolygon q(ndata);

    // convert data to logical (screen) units
    for(int i=0; i<ndata; i++)
       	q.setPoint(i,xax->logic(xp[i],true),yax->logic(yp[i],true));   // in log mode, 0 and less will become log10(MinCgt0)

    if(linePen.width() && ndata>1)
    {
        dc->save();
        dc->setPen(linePen);
	      dc->drawPolyline(q);
        dc->restore();
    }
    if(getsymbol()) {
      if(getsymbol()==12)
        draw2DHist(dc, &q, ndata, r);
      else
        drawSymbol(dc, &q, ndata, fSize);
    }

}

//
// acumulate a 2D histogram of the 'count' points in p (p is in dc coordinates) and plot it on dc
// r is the histogram size
//
void FSDataSet::draw2DHist(QPainter *dc, QPolygon *p, int count, QRect r)
{

  // accumulate the histogram in hist
  long i, j;
  long xsize = r.width();
  long ysize = r.height();
  long area = xsize*ysize;

  if(area==0) return;

  unsigned char * hist = new unsigned char[area];

  for(i=0;i<area;i++) hist[i] = 0;

  p->translate(-r.topLeft()); // move points p so origin is 0

  for(i=0;i<count;i++) {
    long x = p->point(i).x();
    long y = p->point(i).y();
    j = x + y*xsize;
    if( x>0 && x<xsize && j>0 && j<area) {
      hist[j]++;
    }
  }

  QImage image(xsize, ysize, QImage::Format_ARGB32);
  for(i=0;i<area;i++)
    if(hist[i]>symbolSize)
      image.setPixel(i%xsize,i/xsize,qRgb(255-hist[i],0,hist[i]));
    else
      image.setPixel(i%xsize,i/xsize,qRgba(255,255,255,0)); // alpha = 0 so pixels are transparent

  // draw the image on dc
  dc->drawImage(r.adjusted(0,0,-1,-1), image, image.rect().adjusted(0,0,-1,-1));

  delete[] hist;
}

//
// plot on dc the dataset symbol at the count locations p (p is in dc coordinates)
//
void FSDataSet::drawSymbol(QPainter *dc, QPolygon *p, int count, double fSize )
{
  if(symbolSize==0.) return;
  fSize*=symbolSize/FS_SCALE;

  // Set pen for the dc to the value for this line
  dc->save();
  dc->setPen(symbolPen);
  dc->setBrush(symbol%2?QBrush():QBrush(symbolPen.color()));  // odd number are empty, even number are filled

  for(int i=0; i<count; i++)
  {
   int px = p->point(i).x();
   int py = p->point(i).y();

	 switch (symbol) {
	 case 1: // empty circle
	 case 2: // filled circle
		dc->drawEllipse(px-int(fSize),py-int(fSize),int(fSize*2.+.5), int(fSize*2.+.5));
		px--;
		py--;
		break;
	 case 3: // empty box
	 case 4: // filled box
		dc->drawRect(px-int(fSize),py-int(fSize),int(fSize*2.+.5), int(fSize*2.+.5));
		px--;
		py--;
		break;
	 case 5: // empty triangle
	 case 6: // filled triangle
     {
        QPolygon temparray;
        temparray.setPoints(3, px,py-int(fSize*1.15), px-int(fSize),py+int(fSize*0.577), px+int(fSize),py+int(fSize*0.577));
	      dc->drawPolygon(temparray);
     }
		 break;
	 case 7: // reverse empty triangle
	 case 8: // reverse filled triangle
     {
	  	  QPolygon temparray;
        temparray.setPoints(3, px,py+int(fSize*1.15), px-int(fSize),py-int(fSize*0.577), px+int(fSize),py-int(fSize*0.577));
	  	  dc->drawPolygon(temparray);
     }
		 break;
	 case 9: // plus (+)
		dc->drawLine(px,py-int(fSize+.5), px,py+int(fSize+.5));
		dc->drawLine(px-int(fSize+.5),py, px+int(fSize+.5),py);
		break;
	 case 10: // star
		dc->drawLine(px,py+int(fSize), px,py-int(fSize));
		dc->drawLine(px+int(fSize*0.866+.5),py-int(fSize*0.5+.5), px-int(fSize*0.866+.5),py+int(fSize*0.5+.5) );
		dc->drawLine(px-int(fSize*0.866+.5),py-int(fSize*0.5+.5), px+int(fSize*0.866+.5),py+int(fSize*0.5+.5) );
		break;
   case 11: // dot
   default:
    dc->drawPoint(px,py);
	 }
  }
  dc->restore();
}


//
// call FSEditDataSet dialog
//
bool FSDataSet::edit( void )
{
    FSEditDataSet * dsEdit = new FSEditDataSet(this);
    if( dsEdit->exec() )
    {
        legend = dsEdit->legend();
        linePen = dsEdit->linePen();
        symbolPen = dsEdit->symbolPen();
        setsymbol(dsEdit->symbol());
        setsymbolSize(dsEdit->symbolSize());
        return true;
    }
    return false;
}


//
// interpolate linearly the value of yp at xx[i] and set it in yy[i]
// return bitwise OR:
//   0x01 if this class data had to be sorted before interpolation
//   0x02 if some or all xx[i] were where out of xp[i] range (yy[i] set to outVal)
//   0x04 if no data to be interpolated (ndata==0, yy[i] set to outVal)
//   0x08 if no data to interpolate (n==0)
//   0x10 if (xx==NULL or yy==NULL) and n>0 or n<0

// xp[i] are sorted in increasing x order before interpolation, but several consecutive xp[i] with the same value
// are allowed; in that case, the average value of the yp[i,i+1,i+2,...] is used for interpolation
// The xx[i] dont have to be ordered
// Return true if everything's ok (no exception), false otherwise
//
int FSDataSet::interpolate(int n, double xx[], double yy[], double outVal)
{
    int i, returnVal = 0x0;

    if(((xx==NULL || yy==NULL) && n>0) || n<0) returnVal|=0x10;
    if(n==0)returnVal|=0x08;
    if(ndata==0)returnVal|=0x04;
    if(!extrema()) returnVal|=0x01; // calculate extrema and tell if the local data are ordered

    if(returnVal & 0x18) return returnVal; // bad parameters or no data to interpolate, return now
    if(returnVal & 0x04)
    {
        // we have no data in this class: all yy[] = 0.
        for(i=0; i<n; i++) yy[i] = 0.;
        return returnVal;
    }

    // special case where all the values of xp are the same (one or more datapoint)
    // no interpolation possible except for xx[i]==xp[0] for which yy[i] is set to the average of yp[i]
    bool noException = true;
    if(xMin==xMax)
    {
        int i;
        double meanY = yp[0];
        for(i=1; i<ndata; i++) meanY+=yp[i];
        meanY/=ndata;

        for(i=0; i<n; i++)
        {
            if( xx[i]==xp[0] )
                yy[i]=meanY;
            else
            {
                yy[i]=outVal;
                noException = false;
            }
        }
        returnVal|=noException?0x0:0x02;
        return returnVal;
    }

    FSDataSet * copy = new FSDataSet(this);  // we work with a copy of this dataset...
    if(returnVal & 0x01)
        copy->sort(); // ...in case we need to sort it

    // usual case
    for(i=0; i<n; i++)
    {
        if(xx[i]<xMin || xx[i]>xMax)
        {
            // outside extrema
            yy[i] = outVal;   // xx[i] out of range
            noException = false;
        }
        else
        {
            // we interpolate
            for(int j=1; j<ndata; j++)
            {
                if(xx[i]<=copy->xp[j]) // find the first xp>xx
                {

                    // check if there are several xp[i+1,i+2,...]==xp[i] and average their yp[i]
                    double meanYup = copy->yp[j];
                    int number = 1;
                    while(j+number<ndata && copy->xp[j+number]==copy->xp[j])
                    {
                        meanYup+=copy->yp[j+number];
                        number++;
                    }
                    meanYup/=number;

                    // same process for  xp[i-2,i-3,...] relatively to xp[i-1]
                    double meanYdown = copy->yp[j-1];
                    number = 1;
                    while(j-1-number>=0 && copy->xp[j-1-number]==copy->xp[j-1])
                    {
                        meanYdown+=copy->yp[j-1-number];
                        number++;
                    }
                    meanYdown/=number;

                     // interpolate
                    yy[i] = meanYdown+(xx[i]-copy->xp[j-1])*(meanYup-meanYdown)/(copy->xp[j]-copy->xp[j-1]);
                    break; // j loop
                }
            }
        }
    }

    returnVal|=noException?0x0:0x02;
    return returnVal;
}

//
// sort the dataset in increasing order of x using shell sort algorithm
//
void FSDataSet::sort()
{
    int n, h, i, j;
    double tx, ty;

    /* compute largest increment */
    n = ndata;
    h = 1;
    if (n < 14)
        h = 1;
    else if (n > 29524)
        h = 3280;
    else {
        while (h < n)
            h = 3*h + 1;
        h /= 3;
        h /= 3;
    }

    while (h > 0) {

        /* sort-by-insertion in increments of h */
        for (i = h; i < ndata; i++) {
            tx = xp[i];
            ty = yp[i];
            for (j = i-h; j >= 0 && xp[j]>tx; j -= h)
            {
                xp[j+h] = xp[j];
                yp[j+h] = yp[j];
            }
            xp[j+h] = tx;
            yp[j+h] = ty;
        }

        /* compute next increment */
        h /= 3;
    }
}

//
// swap x and y
//
void FSDataSet::swap()
{
    for(int i=0; i<ndata; i++)
    {
        double tmp = xp[i];
        xp[i] = yp[i];
        yp[i] = tmp;
    }
}


bool FSDataSet::savefile(QString fileName, bool saveAsInt)
{
    QString idName = fileName;
    int slash = fileName.lastIndexOf(QRegExp("[\\/\\\\]"));
    if(slash!=-1)
		  idName = fileName.right( fileName.length()-slash-1 );

    QFile file(fileName);
    if (!file.open(QIODevice::WriteOnly | QIODevice::Text)) return false;

    QTextStream f(&file);
	if(saveAsInt) {
		for(int i=0; i<ndata; i++) {
			xp[i] = floor(xp[i]+0.5);
		    yp[i] = floor(yp[i]+0.5);
		}
  } else {
    for(int i=0; i<ndata; i++)
		f << xp[i] << " " << yp[i] << "\n";
  }
    file.close();

	legend = idName;
	dataHaveChanged = false;

    return true;
}


// convolute yp[] data by a Gaussian of width sigma
// WARNING: assumes that xp[] is equaly spaced with interval "slope"
bool FSDataSet::gaussConvol(double FWHMoverSlope) {

    double sigmaOverSlope = FWHMoverSlope/2.35482f; // convert FWHM to sigma
    long i, j, nmax = (int)(5*sigmaOverSlope); // nmax = 5*sigma in terms of channels

    if(sigmaOverSlope <= 0 || nmax <= 0) return false;

    // create one half of a gaussian
    double sum = 1.0;
    double * gauss = new double[nmax];
    gauss[0] = 1.0;
    for( i=1;i<nmax;i++) {
        gauss[i] = exp(-pow((double)i/sigmaOverSlope,2)/2.0);
        sum += 2.0*gauss[i];
    }
    if(sum==0) return false;

    sum = 1.0/sum;
    for(i=0;i<nmax;i++) gauss[i] *= sum; // normalize this half gaussian to 1/2

    double * ypcopy = new double[ndata];
    for(i=0;i<ndata;i++) ypcopy[i] = yp[i]; // create a copy of yp[]

    // carry out convolution
    for(i=0;i<ndata;i++) {
        yp[i] = ypcopy[i]*gauss[0];
        for(j=1;j<nmax;j++) {
            // add both halfs
            if(i+j<ndata) yp[i] += ypcopy[i+j]*gauss[j];
            if(i-j>=0) yp[i] += ypcopy[i-j]*gauss[j];
        }
    }
    delete ypcopy;
    delete gauss;
    return true;
}




