/* This file is part of the Allegria project
   Copyright (C) 2004 Francois Schiettekatte <francois.schiettekatte@umontreal.ca>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

// Dialog that displays a periodic table and that allows single or multiple element selection
// The table can be extended with dummy elements

#include <QMessageBox>
#include <QTextStream>
#include <QFile>
#include <QFileInfo>
#include <QCoreApplication>

#include "PeriodicTable.h"

Atom::Atom() { 
  symbol = ""; 
  Z=0; 
  averageMass = 0.; 
  MAI=0; 
}


int Atom::nIsotopes() { 
  return isotopeMass.count(); 
}

// find MAI mass
void Atom::findMAI() { 
  double maxAbundance=-1.0;
  for(int i=0;i<isotopeAbundance.count();i++)
    if(isotopeAbundance[i]>maxAbundance) {
      maxAbundance=isotopeAbundance[i];
      MAI = i;
    }
}

void Atom::computeAverageMass() {
  averageMass = 0;
  for(int i=0;i<isotopeAbundance.count();i++)
    averageMass += isotopeAbundance[i]*isotopeMass[i].toDouble();
}


PeriodicTable::PeriodicTable(int w, int h, bool singleSelect, QWidget *parent)
              : QWidget( parent )
{
	setWindowTitle("Periodic Table");

  loadAtoms();

  cb.clear();
  for(int i=0; i<atoms.count(); i++) {
    QPushButton * pb = new QPushButton(atoms[i].symbol,this );
		pb->setCheckable(true);
		if(singleSelect) 
			connect( pb, SIGNAL(clicked()), this, SLOT(singleElementSelect()));
		else
			connect( pb, SIGNAL(clicked()), this, SLOT(multiElementSelect()));

	  cb.append(pb);
	}

  createTableOfElements(w,h);
	_selectedElement = atoms[0]; // set neutron as default
}


bool PeriodicTable::loadAtoms() {

	// load elements info (isotope masses, isotopic abundance, symbol) from file isotopes.dat
    QString symbol, mass;
    int Z;
    double abundance;
    Atom at;

    atoms.clear();

    QFile file("data/isotopes.dat");
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
        QMessageBox::warning( this, "Periodic table", QString("Error opening atoms description file 'isotopes.dat'"), "OK" );
      return false;
    }

    // load, computing average mass based on abundance
    QTextStream f(&file);
    for(int i=0;i<4;i++) f.readLine(); // ignore first 4 lines
    while(!f.atEnd()) {
      f>> Z >> mass >> abundance >> symbol;

      if(atoms.count()<=Z) atoms.append(at);
      if(atoms.count()<=Z) {
        QString message = QString("Can't load 'isotopes.dat' because of a discontinuity at element Z=%1 mass=%2 (%3).")
          .arg(Z).arg(mass).arg(symbol);
        QMessageBox::warning( this, "Periodic table", message,"OK");
        file.close();
        return false;
      } 
      atoms[Z].symbol = symbol;
      atoms[Z].Z = Z;
      atoms[Z].isotopeMass.append(mass);
      atoms[Z].isotopeAbundance.append(abundance);
    }
    file.close();

    // find most abundant isotopes
    for(Z=0;Z<atoms.count();Z++) {
      atoms[Z].findMAI();
      atoms[Z].computeAverageMass();
    }
    return true;
}

void PeriodicTable::createTableOfElements(int width, int height)
{
    int x, y;
	  int bwidth = width/18;
	  int bheight = int(height/(9.2));

    // create standard periodic table of elements
    for(int i=0; i<atoms.count(); i++) {

      int Z = atoms[i].Z;

        // y coordinate of element on table
        if(Z<3) y = 0;
        else if(Z<11) y = 1;
        else if(Z<19) y = 2;
        else if(Z<37) y = 3;
        else if(Z<55) y = 4;
        else if(Z>56 && Z<72) y = 7;
        else if(Z<87) y = 5;
        else if(Z>88 && Z<104) y = 8;
        else if(Z>=104) y = 9;
        else y = 6;

        // x coordinate (in button number) of element on table
        switch (y) {
            case 0: x = (Z-1)*17; break;
            case 1: x = (Z- 3) + (Z>  4?10:0); break;
            case 2: x = (Z-11) + (Z> 12?10:0); break;
            case 3: x = (Z-19);                break;
            case 4: x = (Z-37);                break;
            case 5: x = (Z-55) - (Z> 71?14:0); break;
            case 6: x = (Z-87) - (Z>103?14:0); break;
            case 7: x = (Z-57)+2;              break;
            case 8: x = (Z-89)+2;              break;
            case 9: x = (Z-93);                break;
        }

        // position in pixel units
        x *= bwidth; //buttonSide;
        y *= bheight; //buttonSide; 

        // allow space between lanthanide/acinide series and rest of table
        if(y>6*bheight) y += int(bheight*.2);

        // allow space between lanthanide/acinide series and dummy elements
        if(y>9*bheight) y += int(bheight*.4);

        // create the button, a click event will call _buttonClick()
		    cb[i]->move(QPoint(x,y));
		    cb[i]->setFixedSize (bwidth,bheight);
  }
	lastClicked = 0;

    // resize window
	QRect geo = geometry();
	geo.setWidth(width);
	geo.setHeight(height);
	setGeometry(geo);
}

void PeriodicTable::singleElementSelect()
{
	// this fonction should be called only in case of singleSelect
	int i;

	if(lastClicked>0) cb[lastClicked]->setChecked(false); // unclick previous selection
	lastClicked = 0;

    // check which element is selected: button is toggled on
	for (i=1; i<cb.count(); i++) {
		if(cb[i]->isChecked())
				break;
	}

	if(i==cb.count()) // no button clicked!!
		_selectedElement = atoms[0]; // select neutron as default
	else {
		// set selected element to pushed button
		_selectedElement = atoms[i];
		emit elementSelected();
		lastClicked = i;
	}
}

void PeriodicTable::multiElementSelect()
{
		emit elementSelected();
}


Atom PeriodicTable::returnAtom(int atomNum) {
	if(atomNum>=1 || atomNum < atoms.count()) 
		return atoms[atomNum]; 

	return atoms[0]; // return neutron if argument wrong
}


Atom PeriodicTable::returnAtom(QString symbol) {
	for (int i=1; i<atoms.count(); i++) {
		if(symbol.toUpper() == atoms[i].symbol.toUpper() || 
      (atoms[i].symbol.length()==1 && atoms[i].symbol[0].toUpper()==symbol[0].toUpper()  ) )
			return atoms[i];
	}

	return atoms[0];  // return neutron if argument wrong
}

void PeriodicTable::setSelected(int Z, bool sel) {
	if(Z>=1 && Z < atoms.count())
		cb[Z]->setChecked(sel); 

}


// returns a list of selected element in the periodic table.
// Index [0] contains the number of elements in the list
// If no element are selected, function returns NULL.
QList<int> PeriodicTable::multiSelected()
{
  ZSelected.clear();

	// count the number of pressed buttons in element table
	for(int i = 1; i<atoms.count(); i++)
		if(cb[i]->isChecked())
      ZSelected.append(i); 

	return ZSelected;

}

void PeriodicTable::paintEvent( QPaintEvent * ) 
{
	QSize s = size();
	createTableOfElements(s.width(), s.height());

}

